# This file is part of kurchu.
#
# Licensed under the GNU General Public License Version 2
# Fedora-License-Identifier: GPLv2+
# SPDX-2.0-License-Identifier: GPL-2.0+
# SPDX-3.0-License-Identifier: GPL-2.0-or-later
#
# kurchu is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# kurchu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kurchu.  If not, see <https://www.gnu.org/licenses/>.


from pathlib import Path
import logging
import urllib.request

import kurchu.util

log = logging.getLogger("kurchu")


class SourceBase:
	"""
	Base class for sources
	"""

	def __init__(self, source_config: dict, output_dir: Path) -> None:
		self.name = source_config["name"]
		self.url = source_config["url"]
		self.target_subdir = source_config["target"]
		self.sourcetype = source_config["type"]
		self.full_sync = source_config["sync"]
		self.id_file = None
		self.url_to_id = None
		self.source_id = None
		self.output_dir = output_dir

	def __enter__(self):
		return self

	def generate_gather_url(self) -> None:
		if self.source_id is not None:
			gather_url = ""
			if self.url.endswith("/"):
				gather_url = self.url[:-1]
			gather_url = gather_url[: gather_url.rfind("/")]
			gather_url = f"{gather_url}/{self.source_id}/"
			self.url = gather_url

	def download(self) -> None:
		"""Generic method to download sources"""
		self.generate_gather_url()
		if not self.full_sync and self.url_to_id is not None:
			download_targetpath = self.output_dir / "compose" / self.target_subdir
			log.info(f"Creating directory to store source metadata: {download_targetpath}")
			download_targetpath.mkdir(parents=True, exist_ok=True)
			log.info(f"Downloading source metadata from: {self.url_to_id}")
			urllib.request.urlretrieve(self.url_to_id, download_targetpath / self.id_file)
		else:
			self.lftp_mirror(self.name, self.sourcetype, self.url, self.output_dir, self.target_subdir)

	@staticmethod
	def lftp_mirror(name: str, sourcetype: str, url: str, dest: Path, target_subdir: Path) -> None:
		"""Mirror the specified url using lftp"""
		logfile = dest / f"./logs/gather-{sourcetype}-{name}.log"
		logfile.parent.mkdir(parents=True, exist_ok=True)
		mirrordir = dest / f"./compose/{target_subdir}"
		lftp_command = ["lftp", "-c", "mirror", "--verbose", "--continue", "--no-overwrite", "--delete", url, mirrordir]
		kurchu.util.run_command(lftp_command, logfile)



# This file is part of kurchu.
#
# Licensed under the GNU General Public License Version 2
# Fedora-License-Identifier: GPLv2+
# SPDX-2.0-License-Identifier: GPL-2.0+
# SPDX-3.0-License-Identifier: GPL-2.0-or-later
#
# kurchu is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# kurchu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kurchu.  If not, see <https://www.gnu.org/licenses/>.


from pathlib import Path
import json
import logging
import urllib.request


from kurchu.gather.source.base import SourceBase

log = logging.getLogger("kurchu")


class SourceKojiDist(SourceBase):
	"""
	Implements gathering the koji dist-repos source
	"""

	def __init__(self, source_config: dict, output_dir: Path) -> None:
		super().__init__(source_config, output_dir)
		self.id_file = "repo.json"
		self.url_to_id = f"{self.url}/{self.id_file}"
		log.info(f"Identifying source ID for {self.name}")
		kojidistrepo_file = urllib.request.urlopen(self.url_to_id)
		kojidistrepo_dict = json.load(kojidistrepo_file)
		kojidistrepo_file.close()
		self.source_id = kojidistrepo_dict["id"]
		log.info(f"Source ID for {self.name}: {self.source_id}")



# This file is part of kurchu.
#
# Licensed under the GNU General Public License Version 2
# Fedora-License-Identifier: GPLv2+
# SPDX-2.0-License-Identifier: GPL-2.0+
# SPDX-3.0-License-Identifier: GPL-2.0-or-later
#
# kurchu is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# kurchu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kurchu.  If not, see <https://www.gnu.org/licenses/>.


from kurchu.furnish.source.base import SourceBase


class SourceYumRepo(SourceBase):
	"""
	Implements furnishing the YUM repo source
	"""



# This file is part of kurchu.
#
# Licensed under the GNU General Public License Version 2
# Fedora-License-Identifier: GPLv2+
# SPDX-2.0-License-Identifier: GPL-2.0+
# SPDX-3.0-License-Identifier: GPL-2.0-or-later
#
# kurchu is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# kurchu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kurchu.  If not, see <https://www.gnu.org/licenses/>.


from pathlib import Path
import logging

from kurchu.furnish.upload_target.base import UploadTargetBase
import kurchu.util

log = logging.getLogger("kurchu")


class UploadTargetS3(UploadTargetBase):
	"""
	Implements S3 upload target
	"""

	def __init__(self, upload_config: dict, output_dir: Path) -> None:
		super().__init__(upload_config, output_dir)

	def upload(self) -> None:
		"""Upload to S3"""
		if self.generate_directory_index_page:
			log.info(f"Generating directory index pages for {self.compose_id}")
			self.generate_directory_indexes(self.output_dir)
		if self.target_subdir.startswith("/"):
			upload_subdir_path = self.target_subdir[1:]
		else:
			upload_subdir_path = self.target_subdir
		log.info(f"Uploading compose {self.compose_id} to s3://{self.name}/{upload_subdir_path}/")
		logfile = self.output_dir / f"./logs/furnish-{self.uploadtype}-{self.name}.log"
		logfile.parent.mkdir(parents=True, exist_ok=True)
		s3cmd_put_command = [
			"s3cmd",
			"put",
			"--recursive",
			"--acl-public",
			self.output_dir,
			f"s3://{self.name}/{upload_subdir_path}/",
		]
		if not self.public_read:
			s3cmd_put_command.remove("--acl-public")
		kurchu.util.run_command(s3cmd_put_command, logfile)



# This file is part of kurchu.
#
# Licensed under the GNU General Public License Version 2
# Fedora-License-Identifier: GPLv2+
# SPDX-2.0-License-Identifier: GPL-2.0+
# SPDX-3.0-License-Identifier: GPL-2.0-or-later
#
# kurchu is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# kurchu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kurchu.  If not, see <https://www.gnu.org/licenses/>.


from pathlib import Path


class SourceBase:
	"""
	Base class for sources
	"""

	def __init__(self, source_config: dict, output_dir: Path) -> None:
		self.name = source_config["name"]
		self.url = source_config["url"]
		self.target_subdir = source_config["target"]
		self.sourcetype = source_config["type"]
		self.full_sync = source_config["sync"]
		self.id_file = None
		self.path_to_id = None
		self.source_id = None
		self.output_dir = output_dir

	def __enter__(self):
		return self

	def get_source_data(self) -> dict:
		"""Generic method to construct final source dictionary"""
		source_dict = {}
		source_dict["name"] = self.name
		if self.source_id is not None:
			furnish_url = ""
			if self.url.endswith("/"):
				furnish_url = self.url[:-1]
			furnish_url = furnish_url[: furnish_url.rfind("/")]
			furnish_url = f"{furnish_url}/{self.source_id}/"
			source_dict["url"] = furnish_url
		else:
			source_dict["url"] = self.url
		source_dict["target"] = self.target_subdir
		source_dict["type"] = self.sourcetype
		source_dict["sync"] = self.full_sync
		return source_dict



# This file is part of kurchu.
#
# Licensed under the GNU General Public License Version 2
# Fedora-License-Identifier: GPLv2+
# SPDX-2.0-License-Identifier: GPL-2.0+
# SPDX-3.0-License-Identifier: GPL-2.0-or-later
#
# kurchu is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# kurchu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kurchu.  If not, see <https://www.gnu.org/licenses/>.


from pathlib import Path
import os
import subprocess


class UploadTargetBase:
	"""
	Base class for upload targets
	"""

	def __init__(self, upload_config: dict, output_dir: Path) -> None:
		self.name = upload_config["name"]
		self.target_subdir = upload_config["target"]
		self.uploadtype = upload_config["type"]
		self.generate_directory_index_page = upload_config["generate_indexhtml"] or False
		self.public_read = upload_config["public"] or False
		self.compose_id = upload_config["compose_id"]
		self.output_dir = output_dir

	def __enter__(self):
		return self

	def upload(self) -> None:
		"""Base method for upload targets"""
		return None

	@staticmethod
	def generate_directory_indexes(basedir: Path) -> None:
		"""Generate directory index html pages using tree(1)"""
		tree_cmd = [
			"tree",
			"-h",
			"-a",
			"-H",
			".",
			"-L",
			"1",
			"-I",
			"index.html",
			"--noreport",
			"--charset",
			"utf-8",
			"-o",
			"index.html",
		]
		folders = [x[0] for x in os.walk(basedir)]

		for folder in folders:
			if os.path.exists(f"{folder}/index.html"):
				os.remove(f"{folder}/index.html")
			subprocess.run(tree_cmd, cwd=folder)
		return



# This file is part of kurchu.
#
# Licensed under the GNU General Public License Version 2
# Fedora-License-Identifier: GPLv2+
# SPDX-2.0-License-Identifier: GPL-2.0+
# SPDX-3.0-License-Identifier: GPL-2.0-or-later
#
# kurchu is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# kurchu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kurchu.  If not, see <https://www.gnu.org/licenses/>.


from pathlib import Path
import logging

from kurchu.furnish.source import Source as FurnishSource
from kurchu.furnish.upload_target import UploadTarget as FurnishUploadTarget
import kurchu.util

log = logging.getLogger("kurchu")


def run(config: dict, output_dir: Path) -> None:
	"""Furnish the artifacts for consumption"""
	furnish_config = {}
	write_furnished_config = True
	# Store primary metadata
	log.info(f"Furnishing primary metadata for artifacts for {config["compose"]["compose_id"]}")
	furnish_config["compose"] = {}
	furnish_config["compose"]["compose_date"] = config["compose"]["compose_date"]
	furnish_config["compose"]["compose_id"] = config["compose"]["compose_id"]
	furnish_config["compose"]["release_name"] = config["compose"]["release_name"]
	furnish_config["compose"]["release_version"] = config["compose"]["release_version"]
	# Store gathered sources
	if "gather" in config["compose"] and "sources" in config["compose"]["gather"]:
		furnish_config["compose"]["gather"] = {}
		furnish_config["compose"]["gather"]["sources"] = []
		for source_config in config["compose"]["gather"]["sources"]:
			source_object = FurnishSource.new(source_config, output_dir)
			log.info(f"Furnishing source: {source_config['name']}")
			furnish_config["compose"]["gather"]["sources"].append(source_object.get_source_data())
	# Store compiled images
	if "compile" in config["compose"] and "images" in config["compose"]["compile"]:
		furnish_config["compose"]["compile"] = {"images": []}
		for furnished_image in config["compose"]["compile"]["images"]:
			if Path(output_dir / "compose" / furnished_image["target"]).exists():
				log.info(f"Furnishing image: {furnished_image['image_name']}")
				furnish_config["compose"]["compile"]["images"].append(furnished_image)
	# Store furnish settings
	furnish_config["compose"]["furnish"] = {}
	if "furnish" in config["compose"]:
		if config["compose"]["furnish"]["write_compose_info"] is not None:
			write_furnished_config = config["compose"]["furnish"]["write_compose_info"]
	furnish_config["compose"]["furnish"]["write_compose_info"] = write_furnished_config
	# Write the furnished config as json into the artifacts path
	if write_furnished_config:
		Path(output_dir / "compose").mkdir(parents=True, exist_ok=True)
		with open(file=output_dir / "compose" / "kurchu.artifacts.json", mode="wt") as artifacts_json:
			artifacts_json.write(kurchu.util.dict_to_prettyjson(furnish_config))
			log.info(f"Furnished artifacts information as {output_dir / "compose" / "kurchu.artifacts.json"}")
	if "upload_targets" in config["compose"]["furnish"]:
		for upload_target in config["compose"]["furnish"]["upload_targets"]:
			full_upload_target = upload_target | {"compose_id": config["compose"]["compose_id"]}
			upload_object = FurnishUploadTarget.new(full_upload_target, output_dir)
			upload_object.upload()



# This file is part of kurchu.
#
# Licensed under the GNU General Public License Version 2
# Fedora-License-Identifier: GPLv2+
# SPDX-2.0-License-Identifier: GPL-2.0+
# SPDX-3.0-License-Identifier: GPL-2.0-or-later
#
# kurchu is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# kurchu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kurchu.  If not, see <https://www.gnu.org/licenses/>.


from abc import ABCMeta, abstractmethod
from pathlib import Path
import importlib
import logging

from kurchu.util.exceptions import KurchuBuildSysSetupError

log = logging.getLogger("kurchu")


class BuildSys(metaclass=ABCMeta):
	"""
	Factory for BuildSys objects to derive the correct implementation
	"""

	@abstractmethod
	def __init__(self) -> None:
		return None

	@staticmethod
	def new(buildsys_name: str, buildsys_config: dict, output_dir: Path) -> object:
		buildsystype_map = {
			"koji": "Koji",
		}
		try:
			buildsys_class = [
				f"kurchu.compile.buildsys.{buildsys_config[buildsys_name]["type"]}",
				f"BuildSys{buildsystype_map[buildsys_config[buildsys_name]["type"]]}",
			]
			buildsys_obj = importlib.import_module(buildsys_class[0])
			module_name = buildsys_class[1]
			return buildsys_obj.__dict__[module_name](buildsys_name, buildsys_config, output_dir)
		except Exception as issue:
			raise KurchuBuildSysSetupError(
				f"Support for {buildsys_config[buildsys_name]["type"]} not implemented: {issue}"
			)
		log.info(f"Using buildsys backend: {buildsys_class[0]}")



# This file is part of kurchu.
#
# Licensed under the GNU General Public License Version 2
# Fedora-License-Identifier: GPLv2+
# SPDX-2.0-License-Identifier: GPL-2.0+
# SPDX-3.0-License-Identifier: GPL-2.0-or-later
#
# kurchu is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# kurchu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kurchu.  If not, see <https://www.gnu.org/licenses/>.


from pathlib import Path
import logging
import threading

from kurchu.gather.source import Source as GatherSource

log = logging.getLogger("kurchu")


def gather_task(source_config: dict, output_dir: Path) -> None:
	"""Gather the source defined in the configuration"""
	source_object = GatherSource.new(source_config, output_dir)
	log.info(f"Gathering source: {source_config['name']}")
	source_object.download()


def run(sources_config: dict, output_dir: Path) -> None:
	"""Gather the sources defined in the configuration"""
	gather_threads = []
	for source_config in sources_config:
		gt = threading.Thread(target=gather_task, args=(source_config, output_dir))
		gt.start()
		gather_threads.append(gt)
	for gather_thread in gather_threads:
		gather_thread.join()



# This file is part of kurchu.
#
# Licensed under the GNU General Public License Version 2
# Fedora-License-Identifier: GPLv2+
# SPDX-2.0-License-Identifier: GPL-2.0+
# SPDX-3.0-License-Identifier: GPL-2.0-or-later
#
# kurchu is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# kurchu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kurchu.  If not, see <https://www.gnu.org/licenses/>.


from configparser import ConfigParser
from pathlib import Path
import logging

from kurchu.furnish.source.base import SourceBase

log = logging.getLogger("kurchu")


class SourceODCS(SourceBase):
	"""
	Implements furnishing the odcs compose source
	"""

	def __init__(self, source_config: dict, output_dir: Path) -> None:
		super().__init__(source_config, output_dir)
		self.id_file = ".composeinfo"
		self.path_to_id = f"{self.target_subdir}/compose/{self.id_file}"
		log.info(f"Identifying source ID for {self.name}")
		odcscomposeinfo_file = open(file=output_dir / "compose" / self.path_to_id, mode="rt")
		odcscomposeinfo_cfg = ConfigParser()
		odcscomposeinfo_cfg.read_file(odcscomposeinfo_file)
		odcscomposeinfo_file.close()
		self.source_id = odcscomposeinfo_cfg["compose"]["id"]
		log.info(f"Source ID for {self.name}: {self.source_id}")



# This file is part of kurchu.
#
# Licensed under the GNU General Public License Version 2
# Fedora-License-Identifier: GPLv2+
# SPDX-2.0-License-Identifier: GPL-2.0+
# SPDX-3.0-License-Identifier: GPL-2.0-or-later
#
# kurchu is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# kurchu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kurchu.  If not, see <https://www.gnu.org/licenses/>.


import logging

import kurchu.util

from kurchu.compile.buildsys.base import BuildSysBase

log = logging.getLogger("kurchu")


class BuildSysKoji(BuildSysBase):
	"""
	Implements compiling with the Koji build system
	"""

	def build_image(self, image_config: dict) -> None:
		"""Method to build an image using Koji"""
		imagetoolcmd_map = {
			"kiwi": ["kiwi-build", "kiwi_description"],
		}
		koji_basecmd = f"koji --profile={self.extra_data["koji_profile"]}"
		koji_buildbasecmd = (
			f"{koji_basecmd} {imagetoolcmd_map[image_config["image_tool"]][0]} {self.extra_data["koji_build_tag"]}"
		)
		vcs_url = image_config[imagetoolcmd_map[image_config["image_tool"]][1]]["url"]
		vcs_type = image_config[imagetoolcmd_map[image_config["image_tool"]][1]]["vcs"]
		vcs_ref = image_config[imagetoolcmd_map[image_config["image_tool"]][1]]["ref"]
		vcs_rev = self.get_vcs_revision_from_ref(vcs_url, vcs_ref, vcs_type)
		koji_vcsurl = f"{vcs_type}+{vcs_url}#{vcs_rev}"
		koji_buildcmd = f"{koji_buildbasecmd} {koji_vcsurl}"
		if image_config["image_tool"] == "kiwi":
			koji_kiwibuild_params = f"{image_config["kiwi_description"]["path"]} --type={image_config["image_type"]} --kiwi-profile={image_config["variant"]} --release={image_config["image_release"]}"
			if "kiwi_result_bundle_name_format" not in image_config:
				koji_kiwibuild_params += " --result-bundle-name-format=%N-%v-%I.%A"
			else:
				koji_kiwibuild_params += (
					f" --result-bundle-name-format={image_config["kiwi_result_bundle_name_format"]}"
				)
			if "arch" in image_config:
				for buildarch in image_config["arch"]:
					koji_kiwibuild_params += f" --arch={buildarch}"
			if "volid_prefix" in image_config and image_config["image_type"] == "iso":
				koji_kiwibuild_params += f" --type-attr=volid={image_config["volid_prefix"]}-{image_config["release_version"]}-{image_config["image_release"]}"
			if "appid_prefix" in image_config and image_config["image_type"] == "iso":
				koji_kiwibuild_params += f" --type-attr=application_id={image_config["appid_prefix"]}-{image_config["release_version"]}-{image_config["image_release"]}"
			koji_buildcmd = f"{koji_buildcmd} {koji_kiwibuild_params}"
		log.info(
			f"Building {image_config["image_name"]}-{image_config["release_version"]}-{image_config["image_release"]} on {image_config["buildsys"]} Koji instance using {image_config["image_tool"]}"
		)
		logfile = (
			self.output_dir
			/ f"./logs/compile-{image_config["buildsys"]}-{self.buildsystype}-{image_config["image_tool"]}-{image_config["image_name"]}.log"
		)
		logfile.parent.mkdir(parents=True, exist_ok=True)
		buildrun = kurchu.util.run_command(koji_buildcmd.split(), logfile)
		if buildrun.returncode == 0:
			koji_tagbuild_basecmd = f"{koji_basecmd} tag-build"
			for success_tag in self.extra_data["koji_success_tags"]:
				koji_tagbuild_cmd = f"{koji_tagbuild_basecmd} {success_tag} {image_config["image_name"]}-{image_config["release_version"]}-{image_config["image_release"]}"
				kurchu.util.run_command(koji_tagbuild_cmd.split(), logfile)
			self.lftp_mirror(
				image_config["image_name"],
				image_config["buildsys"],
				"koji",
				image_config["gather_url"],
				self.output_dir,
				image_config["target"],
			)



# This file is part of kurchu.
#
# Licensed under the GNU General Public License Version 2
# Fedora-License-Identifier: GPLv2+
# SPDX-2.0-License-Identifier: GPL-2.0+
# SPDX-3.0-License-Identifier: GPL-2.0-or-later
#
# kurchu is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# kurchu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kurchu.  If not, see <https://www.gnu.org/licenses/>.

from urllib.parse import urlparse
from pathlib import Path
from typing import IO

import io
import json
import logging
import subprocess

from kurchu.util.exceptions import KurchuNonexistentPathError

log = logging.getLogger("kurchu")


def dict_to_prettyjson(data: dict) -> str:
	"""Print JSON of dictionary"""
	return json.dumps(data, sort_keys=True, indent=4, separators=(",", ": "))


def validate_url(url: str) -> bool:
	"""Validate URLs"""
	urlresult = urlparse(url)
	return bool(urlresult.scheme and urlresult.netloc)


def run_command(command: list, logfile_path: Path) -> subprocess.CompletedProcess | bool:
	log.info(f"Run command: {command}")
	if not logfile_path.parent.exists():
		raise KurchuNonexistentPathError(
			f"The directory '{logfile_path.parent}' for '{logfile_path.name}' does not exist."
		)
	with open(logfile_path, "w") as logfile:
		return subprocess.run(args=command, stdout=logfile, stderr=subprocess.STDOUT, text=True)


def pipe_to_log(stream: IO[str] | IO[bytes], log_prefix: str) -> None:
	"""Reads lines from a file-like object (bytes or text) and logs them."""
	if isinstance(stream, io.TextIOBase):
		for output_line in iter(stream.readline, ""):
			log.info(f"{log_prefix}: {output_line.rstrip()}")
	elif isinstance(stream, io.BufferedIOBase):
		for output_line in iter(stream.readline, b""):
			try:
				log.info(f"{log_prefix}: {output_line.decode("utf-8").rstrip()}")
			except UnicodeDecodeError:
				log.warning(f"{log_prefix}: Could not decode line: {output_line!r}")
	else:
		log.error(f"{log_prefix}: Unsupported stream type: {type(stream)}")



# This file is part of kurchu.
#
# Licensed under the GNU General Public License Version 2
# Fedora-License-Identifier: GPLv2+
# SPDX-2.0-License-Identifier: GPL-2.0+
# SPDX-3.0-License-Identifier: GPL-2.0-or-later
#
# kurchu is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# kurchu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kurchu.  If not, see <https://www.gnu.org/licenses/>.


from pathlib import Path
import io
import logging
import subprocess
import tempfile

import kurchu.util

log = logging.getLogger("kurchu")


class BuildSysBase:
	"""
	Base class for build systems
	"""

	def __init__(self, buildsys_name, buildsys_config: dict, output_dir: Path) -> None:
		self.name = buildsys_name
		self.buildsystype = buildsys_config[buildsys_name]["type"]
		self.extra_data = buildsys_config[buildsys_name]
		self.output_dir = output_dir

	def __enter__(self):
		return self

	def build_image(self, image_config: dict) -> None:
		"""Generic method to build an image"""
		pass

	@staticmethod
	def get_vcs_revision_from_ref(url: str, ref: str, vcstype: str) -> str | None:
		"""Derive the exact commit/revision from reference/branch/tag"""
		with tempfile.TemporaryDirectory() as vcstmpdir:
			vcstype_map = {"git": [f"git clone {url} {vcstmpdir}", f"git checkout {ref}", "git rev-parse HEAD"]}
			if vcstype not in vcstype_map.keys():
				return None
			log.info(f"Determining revision for {vcstype} repository {url} at reference {ref}")
			for vcs_command in vcstype_map[vcstype][:-1]:
				vcs_process = subprocess.Popen(
					vcs_command.split(), cwd=vcstmpdir, stdout=subprocess.PIPE, stderr=subprocess.PIPE
				)
				stdout_reader = io.TextIOWrapper(vcs_process.stdout, encoding="utf-8", errors="replace")
				stderr_reader = io.TextIOWrapper(vcs_process.stderr, encoding="utf-8", errors="replace")
				while True:
					stdout_line = stdout_reader.readline()
					if stdout_line:
						kurchu.util.pipe_to_log(io.StringIO(stdout_line), vcstype + "-stdout")
					stderr_line = stderr_reader.readline()
					if stderr_line:
						kurchu.util.pipe_to_log(io.StringIO(stderr_line), vcstype + "-stderr")
					if not stdout_line and not stderr_line:
						break
				vcs_proc_exitcode = vcs_process.wait()
				stdout_reader.close()
				stderr_reader.close()
				vcs_process.stdout.close()
				vcs_process.stderr.close()
				if vcs_proc_exitcode != 0:
					log.error(f"{vcstype} command '{vcs_command}' failed with exit code: {vcs_proc_exitcode}")
					return None
			vcs_revproc = subprocess.run(
				vcstype_map[vcstype][-1].split(), cwd=vcstmpdir, text=True, capture_output=True
			)
			vcsrev = vcs_revproc.stdout
			log.info(f"Revision for {vcstype} repository {url} at reference {ref}: {vcsrev}")
			return vcsrev

	@staticmethod
	def lftp_mirror(name: str, buildsysname: str, buildsystype: str, url: str, dest: Path, target_subdir: Path) -> None:
		"""Mirror the specified url using lftp"""
		logfile = dest / f"./logs/compile-{buildsysname}-{buildsystype}-{name}.log"
		logfile.parent.mkdir(parents=True, exist_ok=True)
		mirrordir = dest / f"./compose/{target_subdir}"
		lftp_command = ["lftp", "-c", "mirror", "--verbose", "--continue", "--no-overwrite", "--delete", url, mirrordir]
		kurchu.util.run_command(lftp_command, logfile)



# This file is part of kurchu.
#
# Licensed under the GNU General Public License Version 2
# Fedora-License-Identifier: GPLv2+
# SPDX-2.0-License-Identifier: GPL-2.0+
# SPDX-3.0-License-Identifier: GPL-2.0-or-later
#
# kurchu is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# kurchu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kurchu.  If not, see <https://www.gnu.org/licenses/>.


from pathlib import Path
import datetime
import tomllib


def generate_compose_date(date_format: str) -> str:
	"""Generate a compose date with the specified format"""
	return datetime.datetime.now(datetime.timezone.utc).strftime(date_format)


def read_config(config_filepath: Path) -> dict:
	"""Read the kurchu TOML config and return a dictionary"""
	config_data = {}
	with open(config_filepath, "r") as config_file:
		config_data = tomllib.loads(config_file.read())
	return config_data


def __process_gather_config(config: dict, format_args: dict) -> dict:
	"""Apply all string substitutions to generate configuration for gather section"""
	# Check if gather section exists
	if "gather" not in config["compose"].keys():
		return config
	# Create config dictionary
	final_config = config
	# Finalize sources
	for sources_index in range(len(final_config["compose"]["gather"]["sources"])):
		final_config["compose"]["gather"]["sources"][sources_index]["name"] = final_config["compose"]["gather"][
			"sources"
		][sources_index]["name"].format(**format_args)
		source_format_args = format_args | {"name": final_config["compose"]["gather"]["sources"][sources_index]["name"]}
		final_config["compose"]["gather"]["sources"][sources_index]["url"] = final_config["compose"]["gather"][
			"sources"
		][sources_index]["url"].format(**source_format_args)
		final_config["compose"]["gather"]["sources"][sources_index]["target"] = final_config["compose"]["gather"][
			"sources"
		][sources_index]["target"].format(**source_format_args)
	return final_config


def __process_compile_config(config: dict, format_args: dict) -> dict:
	"""Apply all string substitutions to generate configuration for compile section"""
	# Check if compile section exists
	if "compile" not in config["compose"].keys():
		return config
	# Create config dictionary
	final_config = config
	# Finalize image settings
	final_config["compose"]["compile"]["image_release"] = final_config["compose"]["compile"]["image_release"].format(
		**format_args
	)
	# Populate image targets
	image_format_args = format_args | {
		"base_name": final_config["compose"]["compile"]["base_name"],
		"image_release": final_config["compose"]["compile"]["image_release"],
	}
	for images_index in range(len(final_config["compose"]["compile"]["images"])):
		images_buildsys = final_config["compose"]["compile"]["images"][images_index]["buildsys"]
		final_config["compose"]["compile"]["images"][images_index]["gather_url"] = final_config["compose"]["compile"][
			"buildsys"
		][images_buildsys]["gather_url"]
		image_variant_format_args = image_format_args | {
			"variant": final_config["compose"]["compile"]["images"][images_index]["variant"]
		}
		final_config["compose"]["compile"]["images"][images_index]["image_name"] = "{base_name}-{variant}".format(
			**image_variant_format_args
		)
		image_variant_format_args = image_variant_format_args | {
			"image_name": final_config["compose"]["compile"]["images"][images_index]["image_name"]
		}
		final_config["compose"]["compile"]["images"][images_index]["target"] = final_config["compose"]["compile"][
			"target"
		].format(**image_variant_format_args)
		for images_prop in final_config["compose"]["compile"]["images"][images_index].keys():
			final_config["compose"]["compile"]["images"][images_index][images_prop] = final_config["compose"][
				"compile"
			]["images"][images_index][images_prop].format(**image_variant_format_args)
	# Delete no longer used gather_urls
	for buildsys_prop in final_config["compose"]["compile"]["buildsys"].keys():
		del final_config["compose"]["compile"]["buildsys"][buildsys_prop]["gather_url"]
	# Delete no longer used target
	del final_config["compose"]["compile"]["target"]
	# Finalize image definition configuration
	if "kiwi_result_bundle_name_format" in final_config["compose"]["compile"]["image_tool"]["kiwi"]:
		final_config["compose"]["compile"]["image_tool"]["kiwi"]["kiwi_result_bundle_name_format"].format(
			**image_format_args
		)
	for kiwi_config_prop in final_config["compose"]["compile"]["image_tool"]["kiwi"]["kiwi_description"].keys():
		final_config["compose"]["compile"]["image_tool"]["kiwi"]["kiwi_description"][kiwi_config_prop] = final_config[
			"compose"
		]["compile"]["image_tool"]["kiwi"]["kiwi_description"][kiwi_config_prop].format(**image_format_args)
	# Finalize buildsys
	for buildsys_key in final_config["compose"]["compile"]["buildsys"].keys():
		for buildsys_prop in final_config["compose"]["compile"]["buildsys"][buildsys_key].keys():
			if isinstance(final_config["compose"]["compile"]["buildsys"][buildsys_key][buildsys_prop], list):
				for buildsys_prop_list_index in range(
					len(final_config["compose"]["compile"]["buildsys"][buildsys_key][buildsys_prop])
				):
					final_config["compose"]["compile"]["buildsys"][buildsys_key][buildsys_prop][
						buildsys_prop_list_index
					] = final_config["compose"]["compile"]["buildsys"][buildsys_key][buildsys_prop][
						buildsys_prop_list_index
					].format(**image_format_args)
			else:
				final_config["compose"]["compile"]["buildsys"][buildsys_key][buildsys_prop] = final_config["compose"][
					"compile"
				]["buildsys"][buildsys_key][buildsys_prop].format(**image_format_args)
	return final_config


def __process_furnish_config(config: dict, format_args: dict) -> dict:
	"""Apply all string substitutions to generate configuration for furnish section"""
	# Check if compile section exists
	if "furnish" not in config["compose"].keys():
		return config
	# Create config dictionary
	final_config = config
	# TODO: Revisit once more settings are in place
	return final_config


def process_config(config: dict) -> dict:
	"""Apply all string substitutions to generate final runtime configuration"""
	# Create config dictionary
	final_config = config
	# Set basic values
	final_config["compose"]["compose_date"] = generate_compose_date(
		final_config["compose"]["date_format"] if "date_format" in final_config["compose"].keys() else "%Y%m%d"
	)
	final_config["compose"]["compose_id"] = config["compose"]["compose_id"].format(
		release_version=final_config["compose"]["release_version"], compose_date=final_config["compose"]["compose_date"]
	)
	main_format_args = {
		"release_version": final_config["compose"]["release_version"],
		"release_name": final_config["compose"]["release_name"],
		"compose_date": final_config["compose"]["compose_date"],
		"compose_id": final_config["compose"]["compose_id"],
	}
	# Finalize gather section
	final_config = __process_gather_config(final_config, main_format_args)
	# Finalize compile section
	final_config = __process_compile_config(final_config, main_format_args)
	# Finalize furnish section
	final_config = __process_furnish_config(final_config, main_format_args)
	return final_config



# This file is part of kurchu.
#
# Licensed under the GNU General Public License Version 2
# Fedora-License-Identifier: GPLv2+
# SPDX-2.0-License-Identifier: GPL-2.0+
# SPDX-3.0-License-Identifier: GPL-2.0-or-later
#
# kurchu is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# kurchu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kurchu.  If not, see <https://www.gnu.org/licenses/>.


from pathlib import Path
import logging
import threading

from kurchu.compile.buildsys import BuildSys as CompileBuildSys

log = logging.getLogger("kurchu")


def prepare_imagebuild_cfg(
	imagebuild_config: dict, imagetool_config: dict, release_version: str, image_release: str
) -> dict:
	"""Prepare image build configuration dictionary"""
	imagebuild_config["release_version"] = release_version
	imagebuild_config["image_release"] = image_release
	for imagetool in imagetool_config.keys():
		if imagebuild_config["image_tool"] == imagetool:
			for imagetool_param in imagetool_config[imagetool].keys():
				if imagetool_param not in imagebuild_config:
					imagebuild_config[imagetool_param] = imagetool_config[imagetool][imagetool_param]
	return imagebuild_config


def imagebuild_task(buildsys_object: CompileBuildSys, imagebuild_config: dict) -> None:
	"""Run the image build tasks defined in the configuration"""
	log.info(f"Building image: {imagebuild_config["image_name"]}")
	buildsys_object.build_image(imagebuild_config)


def run(compile_config: dict, release_version: str, output_dir: Path) -> None:
	"""Run the compile step defined in the configuration"""
	buildsys_objects = {}
	for buildsys_cfg in compile_config["buildsys"].keys():
		buildsys_objects[buildsys_cfg] = CompileBuildSys.new(buildsys_cfg, compile_config["buildsys"], output_dir)
	gather_threads = []
	for imagecfg_dict in compile_config["images"]:
		imagebuild_config = prepare_imagebuild_cfg(
			imagecfg_dict, compile_config["image_tool"], release_version, compile_config["image_release"]
		)
		buildsys_object = buildsys_objects[imagebuild_config["buildsys"]]
		gt = threading.Thread(target=imagebuild_task, args=(buildsys_object, imagebuild_config))
		gt.start()
		gather_threads.append(gt)
	for gather_thread in gather_threads:
		gather_thread.join()



# This file is part of kurchu.
#
# Licensed under the GNU General Public License Version 2
# Fedora-License-Identifier: GPLv2+
# SPDX-2.0-License-Identifier: GPL-2.0+
# SPDX-3.0-License-Identifier: GPL-2.0-or-later
#
# kurchu is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# kurchu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kurchu.  If not, see <https://www.gnu.org/licenses/>.


from abc import ABCMeta, abstractmethod
from pathlib import Path
import importlib
import logging

from kurchu.util.exceptions import KurchuUploadTargetSetupError

log = logging.getLogger("kurchu")


class UploadTarget(metaclass=ABCMeta):
	"""
	Factory for UploadTarget objects to derive the correct implementation
	"""

	@abstractmethod
	def __init__(self) -> None:
		return None

	@staticmethod
	def new(upload_config: dict, output_dir: Path) -> object:
		uploadtype_map = {
			"s3": "S3",
		}
		try:
			uploadtarget_class = [
				f"kurchu.furnish.upload_target.{upload_config["type"]}",
				f"UploadTarget{uploadtype_map[upload_config["type"]]}",
			]
			uploadtarget_obj = importlib.import_module(uploadtarget_class[0])
			module_name = uploadtarget_class[1]
			return uploadtarget_obj.__dict__[module_name](upload_config, output_dir)
		except Exception as issue:
			raise KurchuUploadTargetSetupError(f"Support for {upload_config["type"]} not implemented: {issue}")
		log.info(f"Using upload target backend: {uploadtarget_class[0]}")



# This file is part of kurchu.
#
# Licensed under the GNU General Public License Version 2
# Fedora-License-Identifier: GPLv2+
# SPDX-2.0-License-Identifier: GPL-2.0+
# SPDX-3.0-License-Identifier: GPL-2.0-or-later
#
# kurchu is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# kurchu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kurchu.  If not, see <https://www.gnu.org/licenses/>.


from argparse import ArgumentParser
from pathlib import Path
import logging

import kurchu.compile
import kurchu.config
import kurchu.furnish
import kurchu.gather
import kurchu.util

log = logging.getLogger("kurchu")


def cli_main():
	"""Main command-line interface"""
	parser = ArgumentParser(description="gather Fedora/CentOS resources to create artifact collections")
	parser.add_argument("config", type=Path, default="kurchu.toml", help="Path to configuration file written in TOML")
	parser.add_argument("--dest", "-d", type=Path, help="Path to store artifact collection")
	parser.add_argument("--logfile", "-l", type=Path, default=None, help="Override path to log file")
	args = parser.parse_args()

	# Read config from TOML to dictionary
	kurchu_config_dict = kurchu.config.process_config(kurchu.config.read_config(args.config))
	# Set up destination
	base_destdir = args.dest or Path(kurchu_config_dict["compose"].get("destdir", "./artifacts"))
	artifacts_dest = base_destdir / kurchu_config_dict["compose"]["compose_id"]
	# Set up logfile path
	if not args.logfile:
		logfile = artifacts_dest / "logs" / "kurchu.log"
	else:
		logfile = args.logfile
	logfile.parent.mkdir(parents=True, exist_ok=True)
	logging.basicConfig(filename=logfile, encoding="utf-8", level=logging.DEBUG)
	# Print config to the screen
	print(kurchu.util.dict_to_prettyjson(kurchu_config_dict))
	# Gather sources
	if "gather" in kurchu_config_dict["compose"] and "sources" in kurchu_config_dict["compose"]["gather"]:
		print("Gathering sources...")
		kurchu_sources_list = kurchu_config_dict["compose"]["gather"]["sources"]
		kurchu.gather.run(kurchu_sources_list, artifacts_dest)
		print("Completed gathering sources!")
	# Compile images
	if "compile" in kurchu_config_dict["compose"]:
		print("Compiling images...")
		kurchu_compile_dict = kurchu_config_dict["compose"]["compile"]
		kurchu_config_releasever = kurchu_config_dict["compose"]["release_version"]
		kurchu.compile.run(kurchu_compile_dict, kurchu_config_releasever, artifacts_dest)
		print("Completed compiling images!")
	# Furnish artifacts
	kurchu.furnish.run(kurchu_config_dict, artifacts_dest)



# This file is part of kurchu.
#
# Licensed under the GNU General Public License Version 2
# Fedora-License-Identifier: GPLv2+
# SPDX-2.0-License-Identifier: GPL-2.0+
# SPDX-3.0-License-Identifier: GPL-2.0-or-later
#
# kurchu is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# kurchu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kurchu.  If not, see <https://www.gnu.org/licenses/>.


class KurchuError(Exception):
	"""
	Base class for all exceptions

	All specific exceptions are subclasses of KurchuError.
	"""

	def __init__(self, message: str):
		self.message = message

	def __str__(self):
		return format(self.message)


class KurchuNonexistentPathError(KurchuError):
	"""
	Exception raised when attempting open a nonexistent path
	"""


class KurchuSourceSetupError(KurchuError):
	"""
	Exception raised when attempting to instantiate an unsupported
	source.
	"""


class KurchuUploadTargetSetupError(KurchuError):
	"""
	Exception raised when attempting to instantiate an unsupported
	upload target.
	"""


class KurchuBuildSysSetupError(KurchuError):
	"""
	Exception raised when attempting to instantiate an unsupported
	upload target.
	"""
